/**
 * End of line; an empty line ("\n" or "\r\n"). These voids are tracked to allow
 * for spacing groups apart arbitrarily.
 *
 * Only editable in the text editor not the UI.
 *
 * Numerical IDs are used over symbols since they can be serialized. to-do: do I
 * need that? Is that even desirable for un/redo?
 */
export type EOL = {
    readonly end: LineEnding | '';
    readonly id: number;
    readonly type: 'EOL';
    readonly up: Group;
};
/**
 * Top-level array of lines with optional trailing EOLs.
 *
 * A group can be empty in the UI but is parsed out.
 *
 * Doesn't imply indentation.
 */
export type Group = {
    readonly down: (Line | EOL)[];
    readonly id: number;
    readonly type: 'Group';
    readonly up: TextTree;
};
/** Positive integral indentation width in spaces or tab. */
export type IndentW = number | 'Tab';
/**
 * A line of text and any subordinate lines. Lines can have empty text but are
 * parsed out.
 *
 * Indentation is implied by parent.
 */
export type Line = {
    readonly down: Line[];
    /** Line ending. */
    end: LineEnding | '';
    expand: boolean;
    readonly id: number;
    /** Integral indentation level. Spacing is synthesized on serialization. */
    indent: number;
    text: string;
    readonly type: 'Line';
    up: Line | Group;
};
/**
 * Unix newline (\n) or Windows carriage return (\r) + newline.
 *
 * An encoding is used instead of escape sequences so snapshots don't get
 * jumbled.
 */
export type LineEnding = 'CRLF' | 'LF';
export type PositionLine = 'End' | 'In' | 'Start';
/**
 * A text file tree representation.
 *
 * This is a plain object without any methods since the store is expected to be
 * the only caller for mutations and the UI only reads.
 *
 * IDs are used to key the UI since changing an object reference would otherwise
 * disassociate an updated model from its UI.
 */
export type TextTree = {
    readonly down: Group[];
    readonly id: number;
    readonly indentW: IndentW;
    readonly type: 'TextTree';
};
export type IDFactory = () => number;
export declare function IDFactory(): IDFactory;
export declare function TextTree(str: string, indentW: IndentW, id?: IDFactory): TextTree;
export declare namespace TextTree {
    var addLine: (to: Group | Line, at: PositionLine, end: LineEnding, id?: IDFactory) => TextTree;
    var breakLine: (line: Line, at: number, end: LineEnding, id?: IDFactory) => {
        root: TextTree;
        new: Line;
    };
    var moveLine: (from: Line, to: Group | Line, at: PositionLine) => TextTree;
    var removeLine: (line: Line) => TextTree;
    var setLineText: (line: Line, str: string) => TextTree;
    var setLineExpand: (line: Line, expand: boolean) => TextTree;
    var toString: (tree: TextTree) => string;
}
