import { treeAdd, treeRemove, treeRoot, treeTouch } from './tree.js';
export function IDFactory() {
    let id = 0;
    return () => id++; // to-do: this could be a symbol but then can't serialize.
}
const factory = IDFactory();
export function TextTree(str, indentW, id = factory) {
    const tree = { down: [], id: id(), indentW, type: 'TextTree' };
    let prev;
    for (const { start, text, end } of split(str)) {
        if (!text && !start) {
            // EOL. Create group if none exists.
            const group = tree.down.at(-1) ?? treeAdd({ id: id(), type: 'Group' }, tree, -1);
            treeAdd({ id: id(), end, type: 'EOL' }, group, -1);
            continue;
        }
        // Line
        const lineIndent = measureIndent(start, indentW);
        if (lineIndent && tree.down.at(-1)?.down.at(-1)?.type === 'EOL') {
            // Unexpected continuation (indent after EOL). Replace EOL with line as a
            // child of any previous line or group.
            const eol = treeRemove(tree.down.at(-1).down.at(-1));
            prev = treeAdd({
                end: eol.end,
                expand: false,
                id: eol.id,
                indent: prev ? prev.indent + 1 : 0,
                text: '',
                type: 'Line'
            }, prev ?? eol.up, -1);
        }
        if (tree.down.length === 0 ||
            tree.down.at(-1).down.at(-1)?.type === 'EOL') {
            // No group or last line was EOL.
            const group = treeAdd({ id: id(), type: 'Group' }, tree, -1);
            prev = treeAdd({ end, expand: false, id: id(), indent: 0, text, type: 'Line' }, group, -1);
            continue;
        }
        // Parent is existing group / line.
        // Walk up tree by indent.
        while (prev.up.type === 'Line' && lineIndent < prev.indent) {
            prev = prev.up;
        }
        prev = treeAdd({
            end,
            expand: false,
            id: id(),
            indent: lineIndent,
            text,
            type: 'Line'
        }, 
        // Indent matches a previous depth or previous depth was too deep?
        lineIndent <= prev.indent
            ? prev.up // Peer of line.
            : prev, // Child of line.
        -1);
    }
    return tree;
}
TextTree.addLine = (to, at, end, id = factory) => {
    let i;
    let indent;
    let up;
    if (to.type === 'Group') {
        if (at !== 'In')
            throw Error('adding to group requires in position');
        up = to;
        indent = 0;
        i = 0;
    }
    else {
        up = at === 'In' ? to : to.up;
        indent = to.indent + (at === 'In' ? 1 : 0);
        // Hack: when In, i is -1 + 1.
        i = up.down.indexOf(to) + (at === 'Start' ? 0 : 1);
    }
    const line = treeAdd({ end, expand: false, id: id(), indent, text: '', type: 'Line' }, up, i);
    return treeRoot(treeTouch(line));
};
TextTree.breakLine = (line, at, end, id = factory) => {
    if (at < 0 || at > line.text.length) {
        return { root: treeRoot(line), new: line };
    }
    const lhs = line.text.slice(0, at);
    const rhs = line.text.slice(at);
    line.text = lhs;
    line = treeTouch(line);
    const next = treeAdd({
        end,
        expand: false,
        id: id(),
        indent: line.indent,
        text: rhs,
        type: 'Line'
    }, line.up, line.up.down.indexOf(line) + 1);
    return { root: treeRoot(next), new: next };
};
TextTree.moveLine = (from, to, at) => {
    if (from === to)
        return treeRoot(from);
    // to-do: consolidate with textTreeAddLine().
    let i;
    let indent;
    let up;
    if (to.type === 'Group') {
        if (at !== 'In')
            throw Error('adding to group requires in position');
        up = to;
        indent = 0;
        i = 0;
    }
    else {
        up = at === 'In' ? to : to.up;
        indent = to.indent + (at === 'In' ? 1 : 0);
        // Hack: when In, i is -1 + 1.
        i = up.down.indexOf(to) + (at === 'Start' ? 0 : 1);
        // to-do: might need `to.lines.findLastIndex(line => line.type !== 'EOL') + 1` for appending.
    }
    const old = from.up;
    treeRemove(from);
    treeAdd(from, up, i);
    setIndent(from, indent);
    if (to.type === 'Line' && at === 'In')
        to.expand = true;
    if (to.type === 'Line')
        to.end = 'LF'; // to-do: ask for line ending
    // Update old ancestors. old is not under node so node is current.
    treeTouch(old);
    return treeRoot(treeTouch(from));
};
TextTree.removeLine = (line) => {
    treeRemove(line);
    return treeRoot(treeTouch(line.up));
};
TextTree.setLineText = (line, str) => {
    line.text = str.replaceAll(/\r?\n/g, '');
    return treeRoot(treeTouch(line));
};
TextTree.setLineExpand = (line, expand) => {
    line.expand = expand;
    return treeRoot(treeTouch(line));
};
TextTree.toString = (tree) => {
    return nodesToString([tree], tree.indentW);
};
function endToString(end) {
    return end === 'LF' ? '\n' : end === 'CRLF' ? '\r\n' : end;
}
function measureIndent(str, indentW) {
    let indent = 0;
    for (const char of str) {
        // If tabs expected but got spaces, default to 2.
        if (char === ' ')
            indent += 1 / (indentW === 'Tab' ? 2 : indentW);
        else if (char === '\t')
            indent++;
        else
            break;
    }
    return Math.floor(indent);
}
function nodesToString(nodes, indentW) {
    let str = '';
    for (const node of nodes) {
        if (node.type === 'TextTree')
            str += nodesToString(node.down, indentW);
        else if (node.type === 'EOL')
            str += endToString(node.end);
        else if (node.type === 'Group')
            str += nodesToString(node.down, indentW);
        else if (node.type === 'Line') {
            const indent = (indentW === 'Tab' ? '\t' : ' '.repeat(indentW)).repeat(node.indent);
            str += `${indent}${node.text}${endToString(node.end)}${nodesToString(node.down, indentW)}`;
        }
        else
            node;
    }
    return str;
}
function setIndent(line, indent) {
    line.indent = indent;
    for (const sub of line.down)
        setIndent(sub, indent + 1);
}
function* split(str) {
    let start = '';
    let text = '';
    let end = '';
    for (const char of str) {
        if (text) {
            if (char === '\n') {
                if (text.at(-1) === '\r') {
                    end = 'CRLF';
                    text = text.slice(0, -1);
                }
                else
                    end = 'LF';
                yield { start, text, end };
                start = text = end = '';
            }
            else
                text += char;
        }
        else {
            if (char === ' ' || char === '\t')
                start += char;
            else if (char === '\n') {
                end = 'LF';
                yield { start, text, end };
                start = text = end = '';
            }
            else
                text += char;
        }
    }
    if (start || text || end)
        yield { start, text, end };
}
//# sourceMappingURL=text-tree.js.map